// File: vdf_layer.cpp
// Auth: Brian Allen Vanderburg II
// Date: Tuesday, December 16, 2003
// Desc: A layering filter for virtualdub


// Includes

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>


#include "filter.h"
#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"
#include "resource.h"

#include "blndfunc.h"
#include "loadimg.h"

// Prototypes
extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff); 



// Filter data structure
typedef struct _layerData
	{
	int method; // blend method
	int ext1, ext2; // Extra blend methods data
	unsigned int opacity; // opacity of top layer (0 - 256)
	int layer_on_top; // is this layer on top
	Pixel32 bgcolor; // color of this layer
	VBitmap *layer_img; // Image
	char img_file[MAX_PATH]; // Image location
	int x_off; // X and Y offset
	int y_off;

	char ctrl_file[MAX_PATH]; // Image control for the effect
	unsigned char *control; // Effect control
	int ctrl_xoff;
	int ctrl_yoff;
	int def_on;

	unsigned char *blend_data;

	COLORREF custcolor[16];

	IFilterPreview *ifp;
	} layerData;


// Number of blend functions
// (count during module init)
int blend_count = 0;


// Dialog procedure
BOOL CALLBACK layerDialogProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
	// Layer data (Set during WM_INITDIALOG)
	layerData *data = (layerData*)GetWindowLong(hDlg, DWL_USER);

	int pos;
	char buf[100];
	LPDRAWITEMSTRUCT lpdi;
	LOGBRUSH brush;
	HBRUSH hBrush;
	CHOOSECOLOR cc;
	OPENFILENAME ofn;
	char file[MAX_PATH] = "\0";

	static Pixel32 curcolor;

	switch(uMsg)
		{
		case WM_INITDIALOG:
			// Set layer data
			SetWindowLong(hDlg, DWL_USER, lParam);
			data = (layerData*)(lParam);

			// Activate preview button
			data->ifp->InitButton(GetDlgItem(hDlg, IDC_PREVIEW));


			// Set method types
			for(pos = 0; pos < blend_count; pos++)
				{
				SendDlgItemMessage(hDlg, IDC_METHODS, CB_ADDSTRING, 0, (LPARAM)(blendMethods[pos].name));
				}

			// Set control range
			SendDlgItemMessage(hDlg, IDC_OPACITY, TBM_SETRANGE, TRUE, MAKELONG(0, 256));

			// Set the active method
			SendDlgItemMessage(hDlg, IDC_METHODS, CB_SETCURSEL, data->method, 0);
			
			// Layer on top
			if(data->layer_on_top)
				pos = BST_CHECKED;
			else
				pos = BST_UNCHECKED;

			SendDlgItemMessage(hDlg, IDC_ONTOP, BM_SETCHECK, pos, 0);

			// Set opacity
			SendDlgItemMessage(hDlg, IDC_OPACITY, TBM_SETPOS, TRUE, data->opacity);

			wsprintf(buf, "%d", data->opacity);
			SetDlgItemText(hDlg, IDC_OPACITYTEXT, buf);

			// Set color
			curcolor = data->bgcolor;

			wsprintf(buf, "#%02X%02X%02X", (curcolor & 0x00FF0000) >> 16,
				(curcolor & 0x0000FF00) >> 8, (curcolor & 0x000000FF));
			SetDlgItemText(hDlg, IDC_COLORTEXT, buf);

			// Set file name, left and bottom offsets
			SetDlgItemText(hDlg, IDC_LAYERIMG, data->img_file);
			SetDlgItemInt(hDlg, IDC_LEFTOFF, data->x_off, TRUE);
			SetDlgItemInt(hDlg, IDC_BOTTOMOFF, data->y_off, TRUE);

			// Set control image name
			SetDlgItemText(hDlg, IDC_CTRLIMG, data->ctrl_file);
			SetDlgItemInt(hDlg, IDC_CTRLLEFT, data->ctrl_xoff, TRUE);
			SetDlgItemInt(hDlg, IDC_CTRLBOTTOM, data->ctrl_yoff, TRUE);

			// Default control
			SendDlgItemMessage(hDlg, IDC_DEFON, BM_SETCHECK, data->def_on ? BST_CHECKED : BST_UNCHECKED, 0);


			// Set extra descriptions
			if(blendMethods[data->method].ext1)
				SetDlgItemText(hDlg, IDC_EXT1DESC, blendMethods[data->method].ext1);
			else
				SetDlgItemText(hDlg, IDC_EXT1DESC, "NONE");

			if(blendMethods[data->method].ext2)
				SetDlgItemText(hDlg, IDC_EXT2DESC, blendMethods[data->method].ext2);
			else
				SetDlgItemText(hDlg, IDC_EXT2DESC, "NONE");

			// Set extra parameters
			SetDlgItemInt(hDlg, IDC_EXT1, data->ext1, TRUE);
			SetDlgItemInt(hDlg, IDC_EXT2, data->ext2, TRUE);


			return TRUE;

		case WM_DRAWITEM:
			// Draw the color button
			if(wParam == IDC_COLOR)
				{
				lpdi = (LPDRAWITEMSTRUCT)(lParam);

				brush.lbColor = RGB((curcolor & 0x00FF0000) >> 16,
					(curcolor & 0x0000FF00) >> 8, (curcolor & 0x000000FF));
				brush.lbStyle = BS_SOLID;

				hBrush = CreateBrushIndirect(&brush);

				FillRect(lpdi->hDC, &(lpdi->rcItem), hBrush);

				DeleteObject(hBrush);
				}
			break;

		case WM_CLOSE:
			EndDialog(hDlg, 1);
			break;

		case WM_HSCROLL:
			pos = SendDlgItemMessage(hDlg, IDC_OPACITY, TBM_GETPOS, 0, 0);

			wsprintf(buf, "%d", pos);
			SetDlgItemText(hDlg, IDC_OPACITYTEXT, buf);
			break;

		case WM_COMMAND:
			{
			switch(LOWORD(wParam))
				{
				case IDC_METHODS:
					if(HIWORD(wParam) == CBN_SELCHANGE)
						{
						// Update extra parameter descriptions
						pos = SendDlgItemMessage(hDlg, IDC_METHODS, CB_GETCURSEL, 0, 0);

						if(blendMethods[pos].ext1)
							SetDlgItemText(hDlg, IDC_EXT1DESC, blendMethods[pos].ext1);
						else
							SetDlgItemText(hDlg, IDC_EXT1DESC, "NONE");

						if(blendMethods[pos].ext2)
							SetDlgItemText(hDlg, IDC_EXT2DESC, blendMethods[pos].ext2);
						else
							SetDlgItemText(hDlg, IDC_EXT2DESC, "NONE");
						}
					break;

				case IDC_COLOR:
					// Select a color
					cc.lStructSize = sizeof(CHOOSECOLOR);
					cc.hwndOwner = hDlg;
					cc.Flags = CC_RGBINIT | CC_FULLOPEN;
					cc.rgbResult = RGB((curcolor & 0x00FF0000) >> 16,
						(curcolor & 0x0000FF00) >> 8, (curcolor & 0x000000FF));
					cc.lpCustColors = data->custcolor;
	
					if(ChooseColor(&cc))
						{
						// Set current color
						curcolor = (GetRValue(cc.rgbResult) << 16) +
							(GetGValue(cc.rgbResult) << 8) +
							GetBValue(cc.rgbResult);

						wsprintf(buf, "#%02X%02X%02X", GetRValue(cc.rgbResult),
							GetGValue(cc.rgbResult), GetBValue(cc.rgbResult));
						SetDlgItemText(hDlg, IDC_COLORTEXT, buf);

						RedrawWindow((HWND)(lParam), NULL, NULL, RDW_INVALIDATE);
						}
					break;

				case IDC_PREVIEW:
					data->ifp->Toggle(hDlg);
					break;

				case IDCANCEL:
					EndDialog(hDlg, 1);
					break;

				case IDC_LAYERBROWSE:
					ofn.lStructSize = sizeof(ofn);
					ofn.hwndOwner = hDlg;
					ofn.lpstrFilter = "Bitmap Files (*.bmp)\0*.bmp\0\0";
					ofn.lpstrCustomFilter = NULL;
					ofn.nFilterIndex = 1;
					ofn.lpstrFile = file;
					ofn.nMaxFile = MAX_PATH;
					ofn.lpstrFileTitle = NULL;
					ofn.lpstrInitialDir = NULL;
					ofn.lpstrTitle = "Open Image";
					ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
					ofn.lpstrDefExt = NULL;
					ofn.lCustData = 0;
					
					if(GetOpenFileName(&ofn))
						{
						SetDlgItemText(hDlg, IDC_LAYERIMG, file);
						}


					break;

				case IDC_CTRLBROWSE:
					ofn.lStructSize = sizeof(ofn);
					ofn.hwndOwner = hDlg;
					ofn.lpstrFilter = "Bitmap Files (*.bmp)\0*.bmp\0\0";
					ofn.lpstrCustomFilter = NULL;
					ofn.nFilterIndex = 1;
					ofn.lpstrFile = file;
					ofn.nMaxFile = MAX_PATH;
					ofn.lpstrFileTitle = NULL;
					ofn.lpstrInitialDir = NULL;
					ofn.lpstrTitle = "Open Image";
					ofn.Flags = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
					ofn.lpstrDefExt = NULL;
					ofn.lCustData = 0;
					
					if(GetOpenFileName(&ofn))
						{
						SetDlgItemText(hDlg, IDC_CTRLIMG, file);
						}


					break;

				case IDOK:
				case IDC_APPLY:
					// Undo the layer image
					data->ifp->UndoSystem();

					// Get method
					pos = SendDlgItemMessage(hDlg, IDC_METHODS, CB_GETCURSEL, 0, 0);
					if(pos < 0 || pos >= blend_count)
						pos = BLEND_NORMAL;

					data->method = pos;

					// On top or not
					pos = SendDlgItemMessage(hDlg, IDC_ONTOP, BM_GETCHECK, 0, 0);
					if(pos == BST_CHECKED)
						data->layer_on_top = 1;
					else
						data->layer_on_top = 0;

					// Color
					data->bgcolor = curcolor;

					// Opacity
					pos = SendDlgItemMessage(hDlg, IDC_OPACITY, TBM_GETPOS, 0, 0);
					if(pos < 0 || pos > 256)
						pos = 256;
					data->opacity = pos;

					// Get the image file
					GetDlgItemText(hDlg, IDC_LAYERIMG, file, MAX_PATH);
					strcpy(data->img_file, file);

					// Get the control file
					GetDlgItemText(hDlg, IDC_CTRLIMG, file, MAX_PATH);
					strcpy(data->ctrl_file, file);

					// Contro left and bottom offset
					data->ctrl_xoff = GetDlgItemInt(hDlg, IDC_CTRLLEFT, NULL, TRUE);
					data->ctrl_yoff = GetDlgItemInt(hDlg, IDC_CTRLBOTTOM, NULL, TRUE);
					data->def_on = (SendDlgItemMessage(hDlg, IDC_DEFON, BM_GETCHECK, 0, 0) == BST_CHECKED);

					// Get the left offset
					data->x_off = GetDlgItemInt(hDlg, IDC_LEFTOFF, NULL, TRUE);

					// Get the bottom offset
					data->y_off = GetDlgItemInt(hDlg, IDC_BOTTOMOFF, NULL, TRUE);


					// Get extra parameters
					data->ext1 = GetDlgItemInt(hDlg, IDC_EXT1, NULL, TRUE);
					data->ext2 = GetDlgItemInt(hDlg, IDC_EXT2, NULL, TRUE);

					// Rebuild the layer image
					data->ifp->RedoSystem();

					if(LOWORD(wParam) == IDOK)
						EndDialog(hDlg, 0);
					else if(LOWORD(wParam) == IDC_APPLY)
						data->ifp->RedoFrame();

					break;
				}

			break;
			}

		default:
			// Message not handled
			return FALSE;
		}

	// Message handled
	return TRUE;
	}


// runProc
int __cdecl layerRunProc(const FilterActivation *fa, const FilterFunctions *ff)
	{
	PixDim w, h;
	PixDim x, y;
	Pixel32 *src, *dst, *layer;
	Pixel32 top, bottom;
	unsigned char t_r, t_g, t_b;
	unsigned char b_r, b_g, b_b;
	unsigned char d_r, d_g, d_b;
	unsigned char *ctrl;
	layerData *data = (layerData*)(fa->filter_data);

	if(data->method >= blend_count)
		{
		return 1; // Invalid blend type
		}

	ctrl = data->control;

	src = fa->src.data;
	dst = fa->dst.data;
	layer = data->layer_img->data;

	h = fa->src.h;

	y = 0;
	while(h--)
		{
		w = fa->src.w;

		x = 0;
		while(w--)
			{
			// If we are in the effect range, do the effect
			if(*ctrl)
				{
				if(data->layer_on_top)
					{
					top = *layer;
					bottom = *src;
					}
				else
					{
					bottom = *layer;
					top = *src;
					}

				t_r = (unsigned char)((top & 0x00FF0000) >> 16);
				t_g = (unsigned char)((top & 0x0000FF00) >> 8);
				t_b = (unsigned char)(top & 0x000000FF);

				b_r = (unsigned char)((bottom & 0x00FF0000) >> 16);
				b_g = (unsigned char)((bottom & 0x0000FF00) >> 8);
				b_b = (unsigned char)(bottom & 0x000000FF);

				(blendMethods[data->method].func)(t_r, t_g, t_b, b_r, b_g, b_b, &d_r, &d_g, &d_b, x, y, fa->src.w, fa->src.h, fa->pfsi->lCurrentSourceFrame, data->ext1, data->ext2, data->blend_data);

				d_r = ((d_r * data->opacity) + (b_r * (256 - data->opacity))) >> 8;
				d_g = ((d_g * data->opacity) + (b_g * (256 - data->opacity))) >> 8;
				d_b = ((d_b * data->opacity) + (b_b * (256 - data->opacity))) >> 8;

				*dst = (d_r << 16) + (d_g << 8) + d_b;
				}
			else
				{
				// No effect, just copy source pixel
				*dst = *src;
				}


			src++;
			dst++;
			layer++;
			ctrl++;
			x++;
			}

		src = (Pixel32*)((char*)src + fa->src.modulo);
		dst = (Pixel32*)((char*)dst + fa->dst.modulo);
		layer = (Pixel32*)((char*)layer + data->layer_img->modulo);
		y++;
		}

	return 0;
	}

// initProc
int __cdecl layerInitProc(FilterActivation *fa, const FilterFunctions *ff)
	{
	layerData *data = (layerData*)(fa->filter_data);

	data->layer_img = NULL;
	data->control = NULL;

	data->bgcolor = 0x00FFFFFF;
	data->layer_on_top = 1;
	data->method = BLEND_NORMAL;
	data->opacity = 256;

	data->img_file[0] = '\0';
	data->x_off = 0;
	data->y_off = 0;

	data->ctrl_file[0] = '\0';
	data->def_on = 1;
	data->ctrl_xoff = 0;
	data->ctrl_yoff = 0;

	data->blend_data = NULL;

	return 0;
	}

// startProc
int __cdecl layerStartProc(FilterActivation *fa, const FilterFunctions *ff)
	{
	// Build the layer image
	int size;
	VBitmap *tmp;
	layerData *data = (layerData*)(fa->filter_data);

	// Load the layer image
	// The layer size and the source size will be the same
	tmp = LoadImg(&(fa->src), data->bgcolor, data->img_file, data->x_off, data->y_off);
	if(tmp == NULL)
		return 1;

	data->layer_img = tmp;

	// Load the control image
	data->control = (unsigned char*)GlobalAlloc(GMEM_FIXED, fa->src.w * fa->src.h);
	if(data->control == NULL)
		return 1;

	if(LoadCtrl(data->control, data->ctrl_file, fa->src.w, fa->src.h, data->ctrl_xoff, data->ctrl_yoff, data->def_on) == 0)
		return 1;

	// If we need to allocate data for the blend method, do so now
	if(blendMethods[data->method].startfunc)
		{
		// Get data size
		size = blendMethods[data->method].startfunc(NULL, data->ext1, data->ext2, fa->src.w, fa->src.h);

		if(size <= 0)
			return 1;

		// Allocate data
		data->blend_data = (unsigned char*)GlobalAlloc(GMEM_FIXED, size);
		if(data->blend_data == NULL)
			return 1;

		// Try to build lookup table
		return blendMethods[data->method].startfunc(data->blend_data, data->ext1, data->ext2, fa->src.w, fa->src.h);
		}

	return 0;
	}

// endProc
int __cdecl layerEndProc(FilterActivation *fa, const FilterFunctions *ff)
	{
	// Free the layer image
	layerData *data = (layerData*)(fa->filter_data);

	FreeImg(data->layer_img);
	data->layer_img = NULL;

	if(data->control)
		GlobalFree((HGLOBAL)(data->control));

	data->control = NULL;

	if(data->blend_data)
		GlobalFree((HGLOBAL)(data->blend_data));

	data->blend_data = NULL;

	return 0;
	}

// stringProc
void layerStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *buf)
	{
	layerData *data = (layerData*)(fa->filter_data);

	char top[] = "Top";
	char bottom[] = "Bottom";

	char *tmp;

	unsigned char r, g, b;

	if(blend_count == 0)
		return;

	if(data->method >= blend_count)
		return;

	if(data->layer_on_top)
		tmp = top;
	else
		tmp = bottom;

	
	r = (unsigned char)((data->bgcolor & 0x00FF0000) >> 16);
	g = (unsigned char)((data->bgcolor & 0x0000FF00) >> 8);
	b = (unsigned char)(data->bgcolor & 0x000000FF);


	wsprintf(buf, " (%s: %d, %s: %1X%1X%1X)", blendMethods[data->method].name, data->opacity,
		tmp, r, g, b);

	return;
	}

// configProc
int layerConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
	{
	layerData *data = (layerData*)(fa->filter_data);

	data->ifp = fa->ifp;

	return DialogBoxParam(fa->filter->module->hInstModule,
		MAKEINTRESOURCE(IDD_LAYERDLG), hwnd,
		layerDialogProc, (LPARAM)data);
	}

// fssProc for script output
bool layerFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen)
	{
	layerData *data = (layerData*)(fa->filter_data);

	// Parameters Order
	// method
	// ext1, ext2
	// opacity
	// on top
	// color red
	// color green
	// color blue
	// file
	// x and y offset
	// control file
	// control x and y offset
	// default control

	_snprintf(buf, buflen, "Config(%d,%d,%d,%d,%d,%d,%d,%d,\"%s\",%d,%d,\"%s\",%d,%d,%d)",
		data->method, data->ext1, data->ext2,
		data->opacity, data->layer_on_top,
		(data->bgcolor & 0x00FF0000) >> 16,
		(data->bgcolor & 0x0000FF00) >> 8,
		(data->bgcolor & 0x000000FF),
		data->img_file, data->x_off, data->y_off,
		data->ctrl_file, data->ctrl_xoff, data->ctrl_yoff, data->def_on);

	return true;
	}

// script config function
void layerConfigFunc(IScriptInterpreter *isi, void *fa_ptr, CScriptValue *argv, int argc)
	{
	FilterActivation *fa = (FilterActivation*)(fa_ptr);
	layerData *data = (layerData*)(fa->filter_data);

	int r, g, b;

	// Parameters Order
	// method
	// ext1, ext2
	// opacity
	// on top
	// color red
	// color green
	// color blue
	// file
	// x and y offset
	// control file
	// control x and y offset
	// default control



	// Get method
	data->method = argv[0].asInt();
	if(data->method > blend_count)
		data->method = 0;
	if(data->method < 0)
		data->method = 0;

	// Get extra parameters
	data->ext1 = argv[1].asInt();
	data->ext2 = argv[2].asInt();

	// Get opacity
	data->opacity = argv[3].asInt();
	if(data->opacity > 256)
		data->opacity = 256;
	if(data->opacity < 0)
		data->opacity = 0;

	// Get ontop
	data->layer_on_top = argv[4].asInt();

	// Get red
	r = argv[5].asInt();
	if(r > 255)
		r = 255;
	if(r < 0)
		r = 0;

	// Get green
	g = argv[6].asInt();
	if(g > 255)
		g = 255;
	if(g < 0)
		g = 0;

	// Get blue
	b = argv[7].asInt();
	if(b > 255)
		b = 255;
	if(b < 0)
		b = 0;

	// Build color
	data->bgcolor = (r << 16) + (g << 8) + b;

	// Image file
	strncpy(data->img_file, *argv[8].asString(), MAX_PATH);

	// Image file offset
	data->x_off = argv[9].asInt();
	data->y_off = argv[10].asInt();

	// Control file
	strncpy(data->ctrl_file, *argv[11].asString(), MAX_PATH);

	// Control offset
	data->ctrl_xoff = argv[12].asInt();
	data->ctrl_yoff = argv[13].asInt();

	// Default on or off
	data->def_on = argv[14].asInt();


	}


// Parameters Order
// method
// ext1, ext2
// opacity
// on top
// color red
// color green
// color blue
// file
// x and y offset
// control file
// control x and y offset
// default control

// function objects
ScriptFunctionDef layerFuncDefs[] =
	{
		{ (ScriptFunctionPtr)layerConfigFunc, "Config", "0iiiiiiiisiisiii" },
		{ NULL }
	};

/* Script object */
CScriptObject layerObject = 
	{
	NULL, layerFuncDefs, NULL
	};

// Filter definition
struct FilterDefinition layerDef =
	{
	NULL, NULL, NULL,

	"Layer", // name
	"Blends a color layer with the frame.", // desc
	"Brian Allen Vanderburg II", // maker

	NULL, // private_data
	sizeof(layerData), // inst_data_size

	layerInitProc, // initProc
	NULL, // deinitProc
	layerRunProc, // runProc
	NULL, // paramProc
	layerConfigProc, // configProc
	layerStringProc, // stringProc
	layerStartProc, // startProc
	layerEndProc, // endProc

	&layerObject, // script_obj
	layerFssProc // fssProc
	};


// Add filter
static FilterDefinition *fd_layer;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat)
	{
	InitCommonControls();

	blend_count = 0;

	// Count the number of blend functions
	while(blendMethods[blend_count].name)
		blend_count++;

	fd_layer = ff->addFilter(fm, &layerDef, sizeof(FilterDefinition));
	if(fd_layer == NULL)
		return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
    vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

	return 0;
	}

// Remove filter
void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff)
	{
	ff->removeFilter(fd_layer);
	}



BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
	{
	return TRUE;
	}


