// File: win_bmp.cpp
// Auth: Brian Allen Vanderburg II
// Date: Sunday, January, 4, 2004
// Desc: Functions to load and access Windows BMP images


// Includes
#include <windows.h>

#include "win_bmp.h"


// Calculate the used pallete entries
static int CalcUsedPaletteEntries(int bitcount)
	{
	if(bitcount >= 1 && bitcount <= 8)
		return (1 << bitcount);

	return 0;
	}

// Calculate the pitch of a line
int CalcPitch(int line)
	{
	return ((line + 3) & ~3);
	}

// Calculate the size of a line
int CalcLine(int width, int bitcount)
	{
	return ((width * bitcount) + 7) / 8;
	}

// Get the palette
RGBQUAD *CalcPalette(BITMAPINFOHEADER *bih)
	{
	if(bih->biBitCount < 16)
		return (RGBQUAD*)(((unsigned char*)(bih)) + sizeof(BITMAPINFOHEADER));
	else
		return NULL;
	}

// Get the bits
unsigned char *CalcBits(BITMAPINFOHEADER *bih)
	{
	return (((unsigned char*)(bih)) + sizeof(BITMAPINFOHEADER) + (sizeof(RGBQUAD) * CalcUsedPaletteEntries(bih->biBitCount)));
	}

// Calculate the location of a scanline
// zero = the bottom line
unsigned char *CalcScanline(BITMAPINFOHEADER *bih, int line)
	{
	unsigned char *bits;

	// Are we a top-down image
	if(bih->biHeight < 0)
		line = (bih->biHeight - 1 - line);

	bits = CalcBits(bih);

	return (bits + (CalcPitch(CalcLine(bih->biWidth, bih->biBitCount)) * line));
	}

// This function will allocate memory for a DIB
static BITMAPINFOHEADER *AllocImage(int width, int height, int bitcount)
	{
	int dib_size;

	if(height < 0)
		height = - height;

	dib_size = sizeof(BITMAPINFOHEADER);
	dib_size += (sizeof(RGBQUAD) * CalcUsedPaletteEntries(bitcount));
	dib_size += (CalcPitch(CalcLine(width, bitcount)) * height);

	return (BITMAPINFOHEADER*)GlobalAlloc(GMEM_FIXED, dib_size);
	}


// This function will free an image
void FreeBitmap(BITMAPINFOHEADER *bih)
	{
	if(bih)
		GlobalFree((HGLOBAL)(bih));
	}

// This function will load an image
BITMAPINFOHEADER *LoadBitmapFile(char *file)
	{
	HANDLE hFile;
	BITMAPFILEHEADER bfh;
	BITMAPINFOHEADER bih;
	BITMAPINFOHEADER *image;

	DWORD BytesRead;

	// Open the image file
	hFile = CreateFile(file, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, NULL);
	if(hFile == INVALID_HANDLE_VALUE)
		return NULL;

	// Read the file header
	if(ReadFile(hFile, &bfh, sizeof(bfh), &BytesRead, NULL) == FALSE)
		{
		CloseHandle(hFile);
		return NULL;
		}
	
	// Make sure the correct number of bytes was read and the type is correct
	if(BytesRead != sizeof(bfh) || bfh.bfType != 0x4D42)
		{
		CloseHandle(hFile);
		return NULL;
		}

	// Now read the bitmap info header
	if(ReadFile(hFile, &bih, sizeof(bih), &BytesRead, NULL) == FALSE)
		{
		CloseHandle(hFile);
		return NULL;
		}

	// Again, make sure the number of bytes read is correct
	if(BytesRead != sizeof(bih))
		{
		CloseHandle(hFile);
		return NULL;
		}

	// Currently, we can only decode BI_RGB (no compression)
	if(bih.biCompression != BI_RGB)
		{
		CloseHandle(hFile);
		return NULL;
		}

	// Allocate memory for the image
	image = AllocImage(bih.biWidth, bih.biHeight, bih.biBitCount);

	if(image == NULL)
		{
		CloseHandle(hFile);
		return NULL;
		}

	// Copy the bitmap info header into the image
	CopyMemory(image, &bih, sizeof(bih));

	// If it has more than 8 bits, set the colors used to 0 (no color palette entries)
	if(image->biBitCount > 8)
		image->biClrUsed = 0;
	else
		image->biClrUsed = CalcUsedPaletteEntries(image->biBitCount);

	// Read the pallete if there is one
	if(image->biBitCount >= 1 && image->biBitCount <= 8)
		{
		if(ReadFile(hFile, CalcPalette(image),
			CalcUsedPaletteEntries(image->biBitCount) * sizeof(RGBQUAD), &BytesRead, NULL) == FALSE)
			{
			CloseHandle(hFile);
			FreeBitmap(image);
			return NULL;
			}

		if(BytesRead != CalcUsedPaletteEntries(image->biBitCount) * sizeof(RGBQUAD))
			{
			CloseHandle(hFile);
			FreeBitmap(image);
			return NULL;
			}
		}

	// Go to the data bits
	if(SetFilePointer(hFile, bfh.bfOffBits, NULL, FILE_BEGIN) == 0xFFFFFFFF)
		{
		CloseHandle(hFile);
		FreeBitmap(image);
		return NULL;
		}

	// Read the image
	if(ReadFile(hFile, CalcBits(image), CalcPitch(CalcLine(image->biWidth, image->biBitCount)) * image->biHeight,
		&BytesRead, NULL) == FALSE)
		{
		CloseHandle(hFile);
		FreeBitmap(image);
		return NULL;
		}

	// Return the image
	CloseHandle(hFile);
	return image;
	}

// This function will get the color value for a pixel
// Y of 0 is the bottom line
int GetPixelRGB(BITMAPINFOHEADER *bih, int x, int y, unsigned char *r, unsigned char *g, unsigned char *b)
	{
	unsigned char *line;
	unsigned short *wline;
	unsigned long *lline;
	RGBQUAD *pal;
	int h;
	int byte, bit;
	unsigned char value;

	
	if(bih->biHeight < 0)
		h = -(bih->biHeight);
	else
		h = bih->biHeight;

	// Make sure we are in boundary
	if(bih == NULL)
		return 0;

	if(x < 0 || y < 0)
		return 0;

	if(x >= bih->biWidth || y >= h)
		return 0;

	// Get the scanline
	line = CalcScanline(bih, y);


	switch(bih->biBitCount)
		{
		case 1:
			byte = x / 8;
			bit = x % 8;

			pal = CalcPalette(bih);
			if(pal == NULL)
				return 0;

			value = line[byte] & (0x80 >> bit);
			
			if(value)
				{
				*r = pal[1].rgbRed;
				*g = pal[1].rgbGreen;
				*b = pal[1].rgbBlue;
				}
			else
				{
				*r = pal[0].rgbRed;
				*g = pal[0].rgbGreen;
				*b = pal[0].rgbBlue;
				}
			return 1;

		case 4:
			byte = x / 2;

			pal = CalcPalette(bih);
			if(pal == NULL)
				return 0;

			if(x % 2)
				{
				value = line[byte] & 0x0F;
				}
			else
				{
				value = (line[byte] & 0xF0) >> 4;
				}

			*r = pal[value].rgbRed;
			*g = pal[value].rgbGreen;
			*b = pal[value].rgbBlue;

			return 1;

		case 8:
			value = line[x];

			pal = CalcPalette(bih);
			if(pal == NULL)
				{
				*r = value;
				*g = value;
				*b = value;
				}
			else
				{
				*r = pal[value].rgbRed;
				*g = pal[value].rgbGreen;
				*b = pal[value].rgbBlue;
				}

			return 1;

		case 16:
			wline = (unsigned short*)(line + (2 * x));


			*b = ((*wline & 0x7C00) >> 7);
			*g = ((*wline & 0x03E0) >> 2);
			*r = ((*wline & 0x001F) << 3);
			return 1;

		case 24:
			lline = (unsigned long*)(line + (3 * x));

			*b = (unsigned char)(*lline & 0x000000FF);
			*g = (unsigned char)((*lline & 0x0000FF00) >> 8);
			*r = (unsigned char)((*lline & 0x00FF0000) >> 16);
			return 1;

		case 32:
			lline = (unsigned long*)(line + (4 * x));

			*b = (unsigned char)(*lline & 0x000000FF);
			*g = (unsigned char)((*lline & 0x0000FF00) >> 8);
			*r = (unsigned char)((*lline & 0x00FF0000) >> 16);
			return 1;

		default:
			return 0;
		}
	}

// This function will get the gray value for a pixel
// It is just an average of the RGB values
// Y of 0 is the bottom line
int GetPixelGray(BITMAPINFOHEADER *bih, int x, int y, unsigned char *gray)
	{
	unsigned char r, g, b;
	if(GetPixelRGB(bih, x, y, &r, &g, &b))
		{
		*gray = ((r + g + b) / 3);
		return 1;
		}
	else
		return 0;
	}

